//	advance.cpp		1/4/92
//		topological sort of advances in Civilization by MicroProse

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

const MarginWidth		= 5;
const NameWidth		= 17;
const ItemWidth		= MarginWidth + NameWidth + 4;
const MaxCharsList	= 79 - ItemWidth;

static void printMargin(FILE*	fp);

enum AdvType {
	None			= 0x0001,
	Technology	= 0x0002,
	Unit			= 0x0004,
	Improvement	= 0x0008,
	WOW			= 0x0010
};

enum Order		{ Ascending, Descending };

struct Advance;

struct AdvanceRef {
	AdvanceRef(Advance* advance = 0) : advance(advance), next(0) {}

	void	add(Advance* advance);
	void	sort(Order order);
	void	addSorted(AdvanceRef* q, Order order);
	void	print(FILE* fp);

	Advance*	advance;
	AdvanceRef*	next;
};

struct Advance {
	Advance(
		char* 	name = "",
		AdvType	type = None) :

		preds(new AdvanceRef), succs(new AdvanceRef), next(0), nPreds(0),
		expandedPreds(new AdvanceRef), expandedSuccs(new AdvanceRef),
		type(type)
	{
		this->name = new char[strlen(name) + 1];
		strcpy(this->name, name);
	}

	Advance*
	add(
		char*		name,
		AdvType	type = None)
	{
		Advance* p;
		Advance* q;
		for (p = next, q = 0; p; q = p, p = p->next)
			if (!stricmp(p->name, name)) {
				if (p->type == None)
					p->type = type;
				return p;
			}
		Advance* newAdvance = new Advance(name, type);
		nAdvances++;

		if (!q)
			next = newAdvance;
		else
			q->next = newAdvance;
		return newAdvance;
	}

	void
	add(
		Advance*	advance)
	{
		Advance* p;
		Advance* q;
		for (p = next, q = 0; p; q = p, p = p->next)
			;
		if (!q)
			next = advance;
		else
			q->next = advance;
		advance->next = 0;
	}

	void
	sort()
	{
		//	loosely adapted from Wirth, Algorithms + Data Structures = Programs,
		//	pp. 182-189
		// make list of advances with no predecessor

		Advance* p = next;
		next = 0;

		while (p) {
			Advance* q = p;
			p = q->next;

			if (!q->nPreds)
				add(q);
		}

		//	sort rest of advances by "subtracting" zero-predecessor advances
		//	from their successors and adding the successors to the list when
		//	they've had all predecessors "subtracted"

		for (p = next; p; p = p->next)
			for (AdvanceRef* s = p->succs->next; s; s = s->next)
				if (!--s->advance->nPreds)
					add(s->advance);

		//	number them, so we can sort their predecessor and successor lists
		int i;
		for (i = 0, p = next; p; i++, p = p->next)
			p->ordinal = i;
		if (i != nAdvances) {
			puts("Error: list is not partially ordered");
			exit(1);
		}

		expand(&Advance::succs, &Advance::expandedSuccs, Ascending);
		expand(&Advance::preds, &Advance::expandedPreds, Descending);
	}

	void
	expand(
		AdvanceRef*	Advance::*source,
		AdvanceRef*	Advance::*dest,
		Order			order)
	{
		for (Advance* p = next; p; p = p->next) {
			p->expand(source, p->*dest);
			(p->*dest)->sort(order);
		}
	}

	void
	expand(
		AdvanceRef* Advance::*source,
		AdvanceRef*	dest)
	{
		for (AdvanceRef* p = (this->*source)->next; p; p = p->next) {
			dest->add(p->advance);
			p->advance->expand(source, dest);
		}
	}

	void
	print(
		FILE*	fp)
	{
		printMargin(fp);
		fputs("          C I V I L I Z A T I O N    A D V A N C E S\n", fp);
		printMargin(fp);
		fputs("                       Cross Reference\n\n\n", fp);
		printMargin(fp);
		fputs("<- = prerequisites, listed in reverse chronological order\n", fp);
		printMargin(fp);
		fputs("-> = successors, listed in forward chronological order\n\n", fp);

		print(fp, Technology | None, "Advances");
		print(fp, Improvement, "Improvements");
		print(fp, WOW, "Wonders of the World");
		print(fp, Unit, "Units");
	}

	void
	print(
		FILE*		fp,
		AdvType	type,
		char*		title)
	{
		fputc('\n', fp);
		printMargin(fp);
		fprintf(fp, "%s\n", strupr(title));
		strset(title, '-');
		printMargin(fp);
		fprintf(fp, "%s\n\n", title);

		for (Advance* p = next; p; p = p->next) {
			if (!(p->type & type))
				continue;

			printMargin(fp);
			fprintf(fp, "%-*.*s", NameWidth, NameWidth, p->name);

			if (p->expandedPreds->next) {
				fprintf(fp, " <- ");
				p->expandedPreds->print(fp);
				fputc('\n', fp);
				if (p->expandedSuccs->next) {
					fputc('\n', fp);
					printMargin(fp);
					fprintf(fp, "%*.*s", NameWidth, NameWidth, "");
				}
			}

			if (p->expandedSuccs->next) {
				fprintf(fp, " -> ");
				p->expandedSuccs->print(fp);
				fputc('\n', fp);
			}
			fputc('\n', fp);
		}
		fputc('\n', fp);
	}

	char*			name;
	Advance*		next;
	AdvanceRef*	preds;
	AdvanceRef*	succs;
	AdvanceRef*	expandedPreds;
	AdvanceRef*	expandedSuccs;
	int			nPreds;
	int			ordinal;
	AdvType		type;

	static int	nAdvances;
};

int	Advance::nAdvances = 0;

void
AdvanceRef::add(
	Advance* advance)
{
	AdvanceRef* p;
	AdvanceRef* q;
	for (p = next, q = 0; p; q = p, p = p->next)
		if (!stricmp(p->advance->name, advance->name))
			return;
	AdvanceRef* newAdvanceRef = new AdvanceRef(advance);

	if (!q)
		next = newAdvanceRef;
	else
		q->next = newAdvanceRef;
}

void
AdvanceRef::sort(
	Order	order)
{
	AdvanceRef* p = next;
	next = 0;

	while (p) {
		AdvanceRef* q = p;
		p = q->next;
		addSorted(q, order);
	}
}

void
AdvanceRef::addSorted(
	AdvanceRef*	q,
	Order			order)
{
	AdvanceRef*	s = 0;
	AdvanceRef*	r = next;

	while (r &&
			 (order == Ascending && r->advance->ordinal < q->advance->ordinal ||
			  order == Descending && r->advance->ordinal > q->advance->ordinal)
			) {
		s = r;
		r = r->next;
	}

	if (!s)
		next = q;
	else
		s->next = q;
	q->next = r;
}

void
AdvanceRef::print(
	FILE*	fp)
{
	int width = 0;
	for (AdvanceRef* p = next; p; p = p->next) {
		if (width + strlen(p->advance->name) + 2 > MaxCharsList) {
			width = 0;
			fprintf(fp, "\n%*.*s", ItemWidth, ItemWidth, "");
		}
		fprintf(fp, "%s%s", p->advance->name, p->next ? ", " : "");
		width += strlen(p->advance->name) + 2;
	}
}

struct Dependency {
	char*		pred;
	char*		succ;
	AdvType	type;
} dependencies[] = {
	"Alphabet",				"Writing",				Technology,
	"Writing", 				"Diplomat",				Unit,
	"Writing", 				"Library",				Improvement,

	"Alphabet",				"MapMaking",			Technology,
	"MapMaking",			"Trireme",				Unit,
	"MapMaking",			"Lighthouse",			WOW,

	"Alphabet",				"Code of Laws",			Technology,
	"Code of Laws",		"Courthouse",			Improvement,

	"Bronze Working",		"Iron Working",		Technology,
	"Iron Working",		"Legion",				Unit,

	"Bronze Working",		"Currency",				Technology,
	"Currency",				"Marketplace",			Improvement,

	"Writing",				"Literacy",				Technology,
	"Code of Laws",		"Literacy",				Technology,
	"Literacy",				"Great Library",		WOW,

	"Code of Laws",		"Republic",				Technology,
	"Literacy",				"Republic",				Technology,

	"Code of Laws",		"Monarchy",				Technology,
	"Ceremonial Burial",	"Monarchy",				Technology,

	"Currency",				"Trade",					Technology,
	"Code of Laws",		"Trade",					Technology,
	"Trade",					"Caravan",				Unit,

	"Literacy",				"Philosophy",			Technology,
	"Mysticism",			"Philosophy",			Technology,

	"Republic",				"Banking",				Technology,
	"Trade",					"Banking",				Technology,
	"Banking",				"Bank",					Improvement,

	"Monarchy",				"Feudalism",			Technology,
	"Masonry",				"Feudalism",			Technology,

	"Trade",					"Medicine",				Technology,
	"Philosophy",			"Medicine",				Technology,
	"Medicine",				"Shakespeare",			WOW,

	"Philosophy",			"Religion",				Technology,
	"Religion",				"Cathedral",			Improvement,
	"Religion",				"J.S. Bach",			WOW,
	"Religion",				"Michelangelo",		WOW,

	"Philosophy",			"Democracy",			Technology,

	"Medicine",				"Chemistry",			Technology,
	"University",			"Chemistry",			Technology,

	"Chemistry",			"Explosives",			Technology,
	"Gunpowder",			"Explosives",			Technology,

	"Explosives",			"Conscription",		Technology,
	"Republic",				"Conscription",		Technology,
	"Conscription",		"Rifleman",				Unit,

	"Masonry",				"Palace",				Improvement,
	"Masonry",				"City Wall",			Improvement,
	"Masonry",				"Pyramids",				WOW,
	"Masonry",				"Great Wall",			WOW,

	"Masonry",				"Construction",		Technology,
	"Currency",				"Construction",		Technology,
	"Construction",		"Aqueduct",				Improvement,
	"Construction",		"Colosseum",			Improvement,
	"Construction",		"Fortress",				Improvement,

	"Masonry",				"Mathematics",			Technology,
	"Alphabet",				"Mathematics",			Technology,
	"Mathematics",			"Catapult",				Unit,

	"Mathematics",			"University",			Technology,
	"Philosophy",			"University",			Technology,
	"University",			"University (Imp)",	Improvement,

	"Construction",		"Engineering",			Technology,
	"The Wheel",			"Engineering",			Technology,

	"Construction",		"Bridge Building",	Technology,
	"Iron Working",		"Bridge Building",	Technology,

	"University",			"Theory of Gravity",	Technology,
	"Astronomy",			"Theory of Gravity",	Technology,
	"Theory of Gravity",	"Isaac Newton",		WOW,

	"Engineering",			"Invention",			Technology,
	"Literacy",				"Invention",			Technology,

	"Bridge Building",	"Railroad",				Technology,
	"Steam Engine",		"Railroad",				Technology,
	"Railroad",				"Darwin's Voyage",	WOW,

	"Theory of Gravity",	"Atomic Theory",		Technology,
	"Physics",				"Atomic Theory",		Technology,

	"Invention",			"Gunpowder",			Technology,
	"Iron Working",		"Gunpowder",			Technology,
	"Gunpowder",			"Musketeer",			Unit,

	"Railroad",				"Industrialization",	Technology,
	"Banking",				"Industrialization",	Technology,
	"Industrialization",	"Transport",			Unit,
	"Industrialization",	"Factory",				Improvement,
	"Industrialization",	"Women's Suffrage",	WOW,

	"Gunpowder",			"Metallurgy",			Technology,
	"University",			"Metallurgy",			Technology,
	"Metallurgy",			"Cannon",				Unit,

	"Metallurgy",			"Steel",					Technology,
	"Industrialization",	"Steel",					Technology,
	"Steel",					"Battleship",			Unit,

	"Industrialization",	"Corporation",			Technology,

	"Industrialization",	"Communism",			Technology,
	"Philosophy",			"Communism",			Technology,
	"Communism",			"United Nations",		WOW,

	"Steel",					"Automobile",			Technology,
	"Combustion",			"Automobile",			Technology,
	"Automobile",			"Armor",					Unit,

	"Corporation",			"Refining",				Technology,
	"Chemistry",			"Refining",				Technology,
	"Refining",				"Power Plant",			Improvement,

	"Corporation",			"Genetic Eng.",		Technology,
	"Medicine",				"Genetic Eng.",		Technology,
	"Genetic Eng.",		"Cure for Cancer",	WOW,

	"Communism",			"Labor Union",			Technology,
	"Mass Production",	"Labor Union",			Technology,
	"Labor Union",			"Mech Infantry",		Unit,

	"Automobile",			"Mass Production",	Technology,
	"Mass Production",	"Submarine",			Unit,
	"Mass Production",	"Mass Transit",		Improvement,

	"Refining",				"Combustion",			Technology,
	"Explosives",			"Combustion",			Technology,
	"Combustion",			"Cruiser",				Unit,

	"Mass Production",	"Recycling",			Technology,
	"Democracy",			"Recycling",			Technology,
	"Recycling",			"Recycling Center",	Improvement,

	"Mass Production",	"Nuclear Fission",	Technology,
	"Atomic Theory",		"Nuclear Fission",	Technology,
	"Nuclear Fission",	"Manhattan Project",	WOW,

	"Nuclear Fission",	"Nuclear Power",		Technology,
	"Electronics",			"Nuclear Power",		Technology,
	"Nuclear Power",		"Nuclear Plant",		Improvement,

	"Nuclear Power",		"Fusion Power",		Technology,
	"Superconductor",		"Fusion Power",		Technology,

	"Horseback Riding",	"Cavalry",				Unit,

	"Ceremonial Burial",	"Temple",				Improvement,

	"The Wheel",			"Chariot",				Unit,

	"Pottery",				"Granary",				Improvement,
	"Pottery",				"Hanging Garden",		WOW,

	"Horseback Riding",	"Chivalry",				Technology,
	"Feudalism",			"Chivalry",				Technology,
	"Chivalry",				"Knight",				Unit,

	"Ceremonial Burial",	"Mysticism",			Technology,
	"Mysticism",			"Oracle",				WOW,

	"Mysticism",			"Astronomy",			Technology,
	"Mathematics",			"Astronomy",			Technology,
	"Astronomy",			"Copernicus",			WOW,

	"Astronomy",			"Navigation",			Technology,
	"Map Making",			"Navigation",			Technology,
	"Navigation",			"Sail",					Unit,
	"Navigation",			"Magellan",				WOW,

	"Navigation",			"Physics",				Technology,

	"Physics",				"Magnetism",			Technology,
	"Magnetism",			"Frigate",				Unit,

	"Physics",				"Steam Engine",		Technology,
	"Invention",			"Steam Engine",		Technology,
	"Steam Engine",		"Ironclad",				Unit,

	"Physics",				"Flight",				Technology,
	"Combustion",			"Flight",				Technology,
	"Flight",				"Fighter",				Unit,

	"Magnetism",			"Electricity",			Technology,
	"Metallurgy",			"Electricity",			Technology,

	"Flight",				"Advanced Flight",	Technology,
	"Electricity",			"Advanced Flight",	Technology,
	"Advanced Flight",	"Bomber",				Unit,
	"Advanced Flight",	"Carrier",				Unit,

	"Electricity",			"Electronics",			Technology,
	"Electronics",			"Hydro Plant",			Improvement,
	"Electronics",			"Hoover Dam",			WOW,

	"Electronics",			"Computers",			Technology,
	"Computers",			"SETI Program",		WOW,

	"Advanced Flight",	"Rocketry",				Technology,
	"Electronics",			"Rocketry",				Technology,
	"Rocketry",				"Nuclear Unit",		Unit,

	"Rocketry",				"Space Flight",		Technology,
	"Computers",			"Space Flight",		Technology,
	"Space Flight",		"SS Structure",		Improvement,
	"Space Flight",		"Apollo Program",		WOW,

	"Space Flight",		"Plastic",				Technology,
	"Plastic",				"SS Component",		Improvement,

	"Plastic",				"SuperConductor",		Technology,
	"Mass Production",	"SuperConductor",		Technology,
	"SuperConductor",		"SDI Defence",			Improvement,

	"Plastic",				"Robotics",				Technology,
	"Robotics",				"Artillery",			Unit,
	"Robotics",				"Mfg Plant",			Improvement,
	"Robotics",				"SS Module",			Improvement,

	NULL
};

static void
printMargin(
	FILE*	fp)
{
	fprintf(fp, "%*.*s", MarginWidth, MarginWidth, "");
}

void
main()
{
	FILE* fp = fopen("advance.txt", "wt");
	putchar('\n');

	Advance advances;

	printf("Working...");
	for (Dependency* dep = dependencies; dep->pred; dep++) {
		Advance* pred = advances.add(dep->pred);
		Advance* succ = advances.add(dep->succ, dep->type);
		pred->succs->add(succ);
		succ->preds->add(pred);
		succ->nPreds++;
	}
	advances.sort();
	printf("\r            \r");
	advances.print(fp);
}
